/* This file is part of Vcsync
   Copyright (C) 2014, 2019 Sergey Poznyakoff
  
   Vcsync is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   Vcsync is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with Vcsync.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The idea behind this module comes from Loic Dachary's sv_sync_www.c.

   To synchronize web directories on each commit add the following to
   CVSROOT/loginfo:
  
       ALL  /bin/vcsync %{s}

   Make sure to add the following statements to your vcsync.conf file:

       # Detach from the controlling terminal
       detach yes;
       # Sleep two seconds before proceding.
       sleep 2;

   This is neccessary to avoid deadlocking, which occurs when cvs is
   trying to acquire the lock held by the commit process that invoked
   vcsync.

   The following description is taken from sv_sync_www.c:

   <quote>
      For test purpose:
      webcvs -n ' bla bla'
      webcvs -n 'foo/bar bla bla'
      webcvs -n 'f'\''oo bla bla'
      webcvs -n 'f'\''oo - New directory'
      webcvs -n 'bla - Imported sources'

      The idea is that it runs a cvs update -l (to prevent recursion) in the 
      directory where the commit was done. Since the command will be called
      once for each directory where a commit did some action there is no
      need for recursion. In the case of an import command this does not
      hold and a recursion must always be done since there is only one
      call to the script for a whole imported tree (this happens when the
      argument contains the Imported source string).

      The %{s} argument in loginfo invocation is a single argument that lists
      the directory and all the files involved. As a special case if the
      directory was added the file list is replaced by '- New directory'. This
      is lame since adding the files -, New and directory will produce the same
      effect, but it's unlikely. The same applies when a whole source tree is
      imported using cvs import in which case the file list is replaced by
      '- Imported sources'.

      There are three cases to take in account (topdir is the absolute path
      of the directory in which the CVS tree was extracted, subdirectory is
      the directory given in argument):
      - commit that modify the top level directory files
        cd topdir ; cvs update -l 
      - commit that adds a new directory or that import a whole source tree
        cd topdir ; cvs update 'subdirectory'
      - commit that modify files in a subdirectory
        cd topdir/subdirectory ; cvs update -l
   </quote>
   
*/

#include "vcsync.h"
#include "wordsplit.h"
#include <sys/stat.h>
#include <unistd.h>

static int
split_arg(char *arg, char **pdir)
{
	char *p;
	
	if ((p = strchr(arg, ' '))) {
		if (p == arg)
			*pdir = NULL;
		else {
			*pdir = arg;
			*p = 0;
		}
		++p;

		debug(3, ("cvs: dir=%s, file list=%s", arg, p));
		/* If this is a new directory or an import command,
		   then there are no files to collect in the argument list. */
		if (strcmp(p, "- New directory") == 0
		    || strcmp(p, "- Imported sources") == 0)
			return 0;
		
		return 1;
	}
	*pdir = NULL;
	return -1;
}

static void
readinput(void)
{
	char *inbuf = NULL;
	size_t insize = 0;
	while (grecs_getline(&inbuf, &insize, stdin) > 0) {
		size_t len = strlen(inbuf);

		if (len == 0)
			continue;
		if (inbuf[len-1] == '\n')
			inbuf[--len] = 0;
		
		debug(1, ("CVS: %s", inbuf));
	}
	free(inbuf);
}

char *
cvs_proc(int argc, char **argv)
{
	int files;
	char *directory;
	struct wordsplit ws;
	char *cmd = NULL;
	size_t cmdsize;
	char *repo_root;

	if (!argc) 
		die(EX_USAGE, "cvs: expected at least one argument");
	else if (argc == 1) {
		debug(1, ("cvs: (old style?) %s", argv[0]));
		files = split_arg(argv[0], &directory);
		if (files == -1)
			die(EX_FAIL, "cvs: cannot split argument %s", argv[0]);
	} else {
		directory = argv[0];
		files = argc > 1;
	}		
	debug(2, ("cvs: directory=%s, has files=%d", directory, files));

	readinput();

	if (strcmp(directory, "CVSROOT") == 0) {
		diag(LOG_NOTICE, "cvs: ignoring %s", directory);
		return NULL;
	}
	
	repo_root = getenv("CVSROOT");
	if (!repo_root)
		die(EX_FAIL, "CVSROOT not defined");
	debug(2, ("CVSROOT=%s",repo_root));
	/* Set up VCS-dependent defaults */
	if (!config->vcs_command)
		config->vcs_command = "/usr/bin/cvs -q -z3";

	if (!dry_run && directory && files) {
		struct stat st;
		char *dir = mkfilename(repo_root, directory);
		if (stat(dir, &st))
			die(EX_FAIL, "cannot stat repository %s: %s",
			    dir, strerror(errno));
		if (!S_ISDIR(st.st_mode))
			die(EX_FAIL, "%s: not a directory", dir);
		free(dir);
	}
	
	if (chdir(config->dest_root))
		die(EX_FAIL, "cannot change to %s: %s",
		    config->dest_root, strerror(errno));

	if (files && directory && access(directory, F_OK)) {
		if (errno == ENOENT) {
			diag(LOG_INFO,
			     "%s does not exist: attempting clean checkout",
			     directory);
			files = 0;
		} else {
			die(EX_FAIL, "cannot access directory %s", directory,
			    strerror(errno));
		}
	}

	debug(1, ("updating %s/%s from CVS", config->dest_root,
		  directory ? directory : ""));
	if (directory) {
		if (files) {
			if (chdir(directory))
				die(EX_FAIL, "cannot change to %s: %s",
				    directory, strerror(errno));
			grecs_asprintf(&cmd, &cmdsize,
				       "%s update -l", config->vcs_command);
		} else
			grecs_asprintf(&cmd, &cmdsize,
				       "%s -d %s checkout %s",
				       config->vcs_command,
				       repo_root, directory);
	} else if (files)
		grecs_asprintf(&cmd, &cmdsize, "%s update -l",
			       config->vcs_command);
	else
		die(EX_USAGE, "unexpected files without a directory");

	if (wordsplit(cmd, &ws, WRDSF_DEFFLAGS))
		die(EX_FAIL, "wordsplit: %s", wordsplit_strerror(&ws));

	runcmd(ws.ws_wordv, NULL);

	wordsplit_free(&ws);

	if (directory)
		register_directory(directory, !files);
	return NULL;
}

